functor Repl (
  val language_name : string
  val evaluation_style : string
  type exp
  val subst : string * exp -> exp -> exp
  val pp_exp : exp -> string
  val step : exp -> exp option
  val parse : string -> exp
) = struct
  fun subst_ctx ctx e = foldl (fn ((y, ee), e') => subst (y, ee) e') e ctx
  fun println s = print (s ^ "\n")

  local
    fun drop_while f = (fn
      [] => []
    | x::xs => if f x then drop_while f xs else x::xs
    )
    fun take_while f = (fn
      [] => []
    | x::xs => if f x then x::take_while f xs else []
    )
    exception NotDeclaration
    fun parse_dec s =
      if String.isPrefix "val" s
      then let
        val cs = explode s
        val after_val = List.drop (cs, 3)
        val id_then_rest = drop_while Char.isSpace after_val
        val id = implode (take_while Char.isAlphaNum id_then_rest)
        val _::exp_str =
          drop_while (fn #"=" => false | _ => true) id_then_rest
        val exp = parse (implode exp_str)
      in
        (id, exp)
      end else raise NotDeclaration
  in
    val parse_declaration = (fn s =>
      SOME (parse_dec s) handle _ => NONE
    )
  end

  fun repl interactive =
    let
      fun eval e =
        case step e of
          NONE =>
            if interactive
            then (
              println "Evaluation complete";
              e
            ) else e
        | SOME e' =>
            if interactive
            then (
              print "==> ";
              print (pp_exp e');
              TextIO.inputLine TextIO.stdIn;
              eval e'
            ) else eval e'

      fun loop ctx = (
        print "> ";
        let val s_opt = TextIO.inputLine TextIO.stdIn in
          case s_opt of
            (NONE | SOME "exit\n" | SOME "exit;\n") => ()
          | SOME "\n" => loop ctx
          | SOME s =>
            let
              val s' = substring (s, 0, size s - 1)
              val s' = if String.isSuffix ";" s' then substring (s', 0, size s' - 1) else s'
            in
              case parse_declaration s' of
                NONE => (
                  let
                    val e = parse s'
                    val e' = eval (subst_ctx ctx e)
                  in
                    println ("evaluates to " ^ pp_exp e');
                    loop ctx
                  end handle err => (
                    println ("Something went wrong: " ^ exnMessage err);
                    loop ctx
                  )
                )
              | SOME (x, e) =>
                let val e' = eval (subst_ctx ctx e) in
                  println ("binding " ^ x ^ " to " ^ pp_exp e');
                  loop ((x, e')::ctx)
                end
            end
        end
      )
    in
      println ("Welcome to the " ^ language_name ^ " repl.");
      println ("This uses " ^ evaluation_style ^ " dynamic semantics.");
      (if interactive
        then (
          println "You are in interactive mode;";
          println "expressions you enter will evaluate step-by-step.";
          println "Hit <Enter> to progress to the next step while evaluating."
        ) else (
          println "You are not in interactive mode,";
          println " so expressions you enter will evaluate immediately."
        )
      );
      println "Enter an expression or a declaration of the form";
      println "  val <identifier> = <expression>";
      println "Enter \"exit\" to exit to the SML repl.";
      loop []
    end
end

structure Top = 
struct
  structure LPPRepl = Repl(open LambdaPlusPlus open LambdaPlusPlusParser)
  structure LCRepl = Repl(open LambdaCalculus open LambdaCalculusParser)
  structure TranslateRepl = Repl(
    open LambdaCalculus
    val language_name = "Lambda Plus Plus to lambda calculus"
    val parse = Translate.translate o LambdaPlusPlusParser.parse
  )

  val lpp_repl = LPPRepl.repl
  val lc_repl = LCRepl.repl
  val translate_repl = TranslateRepl.repl
end